import argparse
import json
import csv
import sys
import logging
import os
import datetime
from pathlib import Path
import shutil
import InputMain as im
import Units as u
import multiprocessing as mp

# Do not delete these imports -- even though pycharm might mark them as unused, they are important
# for instantiating the emitters in loadEmitterFile

# yk 2020-11-17 commented out for "termodynamics lookup only" usecase
# from Well import Well
# from Leak import Leak, CompressorLeak
# from Compressor import Compressor
# from Dehydrator import Dehydrator
# from Pneumatic import Pneumatic
# from IntermittentPneumatic import IntermittentPneumatic
# from ChemicalInjectionPump import ChemicalInjectionPump
# from TankFlash import TankFlash
# from Unloading import Unloading
# from Completion import Completion
# from TestEmitter import TestEmitter
# from LDAR import LDAR
# from OtherEquipment import Other
import inspect


DEFAULT_CONFIG = "config/defaultConfig.json"
DEFAULT_PICKER = "config/pickerConfig.json"
DEFAULT_SCENARIO = "config/scenarioConfig.json"
LOGFORMATSTRING = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
DEFAULT_TS_FMT = "%y%m%d_%H%M%S"


def getArgs(defaultConfig):

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # yk 2020-11-17 dropped options unrelated to thermo lookup
    parser.add_argument("-c", "--config", default=None, help="scenarioInstanceCconfig file")
    # parser.add_argument("-t", "--timestamp", default=None, help="simulation timestamp")
    # parser.add_argument('-r', '--runNumber', default=0, help="scenario number", type=int)
    # parser.add_argument('-i', '--iterate', help="in Drilldown, iterate through all scenarios", action='store_true', default=False)
    # parser.add_argument('-g', '--graphs', help="suppress graphs", action='store_false', default=True)
    parser.add_argument("-f", "--intakeFile", help="overwrites intake excel file in configuration file")
    parser.add_argument("-o", "--output", help="overwrites output dir in configuration file")
    # parser.add_argument("-m", "--mcRuns", help="number of Monte Carlo runs", type=int, default=None)
    parser.add_argument("-p", "--numProcs", help="number of processors for multiprocessing", type=int, default=None)
    # parser.add_argument("-d", "--drilldownUnits", default='DAYS', help="serializer granularity units")
    # parser.add_argument("-x", "--howManyDays", default=30, help="length of serializer run, days", type=int)
    # parser.add_argument("-s", "--start", default=None, help="start serialization time, in sim units", type=float)
    # parser.add_argument("-e", "--end", default=None, help="end serialization time, in sim units", type=float)

    args = parser.parse_args()
    return args

def getBaseParams(args, inputDir = None):
    ret = {}
    if args.intakeFile:
        if inputDir is not None and args.intakeFile.startswith(inputDir):
            # let user include path to intake file
            ret['facilityEquipTmpl'] = f"{args.intakeFile}"
        else:
            dir = "{inputDir}"
            ret["facilityEquipTmpl"] = f"{dir}/{args.intakeFile}"
    if args.output:
        ret['outputDir'] = args.output
    # if args.mcRuns:
    #     ret['monteCarloIterations'] = args.mcRuns
    if args.numProcs:
        ret['processorCount'] = args.numProcs
    # if args.graphs:
    #     ret['graphs'] = args.graphs
    if 'timestamp' in args and args.timestamp:
        ret['scenarioTimestamp'] = args.timestamp
    return ret

def getConfig(defaultConfig=DEFAULT_CONFIG):
    args = getArgs(defaultConfig)

    with open(DEFAULT_PICKER, "r") as cf:       # first config file
        PICKERconfig = json.load(cf)

    with open(DEFAULT_SCENARIO, "r") as cf:     # second config file
        SCENARIOconfig = json.load(cf)

    with open(DEFAULT_CONFIG, "r") as cf:       # third config file
        DEFAULTconfig = json.load(cf)

    config = {**PICKERconfig, **SCENARIOconfig, **DEFAULTconfig}  # concatenate the config files

    config = {**config, **getBaseParams(args, inputDir=config.get('inputDir', None))}  # getBaseParams is here so command line arguments (in particular, inputDir) are available to getGlobals
    gbl = im.Globals()
    globalParams = gbl.getGlobals(config)       # inhale spreadsheet params
    config = {**config, **globalParams, **getBaseParams(args, inputDir=config.get('inputDir', None))}  # getBaseParms is here so command line arguments override spreadsheet args

    if 'scenarioTimestamp' not in config:
        scenarioTimestamp = datetime.datetime.now().strftime(config.get('scenarioTimestampFormat', DEFAULT_TS_FMT))
        config = {**config, 'scenarioTimestamp': scenarioTimestamp}

    return config, args

def readJsonEmitters(path, cls, numLines):
    emitters = []
    if os.path.exists(path):
        with open(path, "r") as iFile:
            rowCount = 0
            for singleLine in iFile:
                rowCount += 1
                if rowCount > numLines:
                    break
                jsonRep = json.loads(singleLine)
                singleEmitter = cls.fromJSON(jsonRep)
                # production sites pseudo equipments (i.e., Unloading and Completion)
                # may not instantiate object if particular well wouldn't have unloading for example.
                # so we will filter those null objects out
                if singleEmitter:
                    emitters.append(singleEmitter)
    return emitters

def loadEmitterFile(inputFile, clsName, numLines=None):
    if clsName not in globals():
        logging.warning(f"{clsName} not found in globals")
        return []
    if numLines is None:
        numLines = sys.maxsize
    cls = globals()[clsName]
    emitters = []
    inputPath = Path(inputFile)
    if inputPath.suffix == '.json':
        emitters = readJsonEmitters(inputPath, cls, numLines)
    else:
        with open(inputFile, "r") as iFile:
            dr = csv.DictReader(iFile)
            rowCount = 0
            for singleLine in dr:
                rowCount += 1
                if rowCount > numLines:
                    break
                singleEmitter = cls.fromCSVRow(**singleLine)
                # production sites' pseudo Emitter may not instantiate when it is not needed (e.g. Well without Unloading)
                # need to skim those null object
                if emitters:
                    emitters.append(singleEmitter)
    return emitters

def getEmitters(config):
    eStructs = config['emitters']
    ret = []
    for singleE in eStructs:
        theseEmitters = loadEmitterFile(singleE['inputFile'], singleE['class'], singleE.get('numInputRows', None))
        ret.extend(theseEmitters)
    return ret

def writeEmitter(file, emitter):
    def myconverter(o):
        if isinstance(o, datetime.datetime):
            return o.strftime('%Y-%m-%d')
        elif isinstance(o, datetime.time):
            return o.strftime('%H:%M:%S')
    str = json.dumps(emitter, default=myconverter)
    file.write(str)
    file.write("\n")

def streamEmitters(file):
    for singleLine in file:
        yield json.loads(singleLine)

def getAgents(config):
    aStructs = config.get('agents', [])
    ret = []
    for singleA in aStructs:
        theseEmitters = loadEmitterFile(singleA['inputFile'], singleA['class'], singleA.get('numInputRows', None))
        ret.extend(theseEmitters)
    return ret

def ensureDirectory(filePath):
    fPath = Path(filePath)
    parentPath = fPath.parent
    parentPath.mkdir(parents=True, exist_ok=True)

def globalInitializer(config, main=False):
    if main:
        logfile = "{}/log_main_{}.log".format(config['LogDir'], os.getpid())
        ensureDirectory(logfile)
        h = [logging.FileHandler(logfile), logging.StreamHandler()]
    else:
        logfile = "{}/log_{}.log".format(config['LogDir'], os.getpid())
        ensureDirectory(logfile)
        h = [logging.FileHandler(logfile)]
    logging.basicConfig(level=logging.DEBUG, format=LOGFORMATSTRING, handlers=h)

def expandTemplate(fStr, config):
    oPath = fStr.format(**config)
    return oPath

def expandFilename(fStr, config):
    oPath = expandTemplate(fStr, config)
    ensureDirectory(oPath)
    return oPath

def expandCopy(iStr, oStr, config):
    iPath = expandFilename(iStr, config)
    oPath = expandFilename(oStr, config)
    shutil.copy(iPath, oPath)

def getSimDuration():
    config, _ = getConfig()
    simDuration = u.daysToSecs(config['testIntervalDays'] + 1)
    return simDuration

def readGCFingerprintFile(config):
    filename = expandFilename(config['gcFingerprintFile'], config)
    gcProfiles = {}
    with open(filename, "r") as iFile:
        for inLine in iFile:
            profileJson = json.loads(inLine)
            gcProfiles[profileJson['profile_id']] = profileJson
    return gcProfiles

def runJobSingleThreaded(fn, argGen, timer=None, limit=sys.maxsize):
        res = []
        count = 0
        for args in argGen:
            res.append(fn(*args))
            count += 1
            if timer:
                timer.incr()
            if count > limit:
                break
        logging.info(f"runJob (singlethread) complete, reslen: {len(res)}")
        return res

def runJobMultiThreaded(fn, argGen, numProcs=1, init=None, initArgs=None, limit=sys.maxsize, timer=None):
    with mp.Pool(processes=numProcs, initializer=init, initargs=initArgs) as p:
        jobs = []
        logging.info("Spawning jobs")
        count = 0
        for args in argGen:
            job = p.apply_async(fn, args)
            jobs.append(job)
            count += 1
            if count > limit:
                break
        logging.info("Awaiting results")
        res = []
        for j in jobs:
            jobRet = j.get()
            logging.debug(f"Job: {j}, result: {jobRet}")
            if timer:
                timer.incr()
            res.append(jobRet)
    logging.info(f"runJob complete, reslen: {len(res)}: numProcs: {numProcs}")
    return res

def runJob(fn, argGen, numProcs=1, init=None, initArgs=None, production=False, limit=sys.maxsize, timer=None):
    logging.info(f"runJob: fn: {fn}, numProcs: {numProcs}")
    if numProcs == 1 and not production:
        return runJobSingleThreaded(fn, argGen, timer, limit)
    return runJobMultiThreaded(fn, argGen, numProcs, init, initArgs, limit, timer)

