# This file reads an Excel workbook that describes a particular site and returns the intake data dictionaries required
# to run the model pickers to instantiate equipment models
import AppUtils as au
import pandas as pd
import Units as u
import json
import copy
from pathlib import Path
import datetime
import numbers
import re

# This is the default state transition matrix that is used if not specified by the user
DEFAULT_XITION_PROB_MATRIX = {"OPERATING": {"NOT_OPERATING_PRESSURIZED": 0.2, "BLOWDOWN": 0.8},
                              "NOT_OPERATING_PRESSURIZED": {"STARTING": 0.2, "BLOWDOWN": 0.8},
                              "NOT_OPERATING_DEPRESSURIZED": {"STARTING": 1.0, "NOT_OPERATING_PRESSURIZED": 0.0}
                              }

# These are the default timing parameters used for compressor states if not specified by the user. They are given as
# triangular distributions with units of hours
DEFAULT_TIMING = {"opTime": {"distribution": "triangular", "min": 1, "mean": 420, "max": 720},
                  "nopTime": {"distribution": "triangular", "min": 0.5, "mean": 8, "max": 24},
                  "nodTime": {"distribution": "triangular", "min": 0.5, "max": 48}
                  }
# Makes a singleton pandas dataframe that contains all parameters from "Global_Simulation_Parameters" sheet of workbook
class Globals:

    GlobalSingleton = {}
    GlobalRenames = {
        'Simulation Start Date': 'simulationStartDate',
        'Simulation Duration [Days]': 'testIntervalDays',
        'Plot Interval': 'plotInterval',
        'Number of Monte Carlo Iterations': 'monteCarloIterations',
        'Output Directory': 'outputDir',
        'Tank Flash Save Interval Threshold [min]': 'tankFlashSaveIntervalThresholdMin',
    }

    # let user to use both string and Excel Date to sepcify start date
    # .__func__ is needed as handleDate is a staticmethod
    # https://stackoverflow.com/questions/41921255/staticmethod-object-is-not-callable/41921291#41921291
    @staticmethod
    def handleDate(v):
        if isinstance(v, (pd.Timestamp, datetime.datetime)):
            return v.strftime('%Y-%m-%d')
        elif isinstance(v, str):
            return v
        else:
            raise TypeError('Globals.handleDate() expects datetime or string, got: value = %s, type = %s' % (str(v), str(type(v))))
    GlobalFuncs = {
        'simulationStartDate': handleDate.__func__
    }

    def getGlobals(self, config, configDir=Path('config'), userConfigFile='userConfig.json'):
        if not Globals.GlobalSingleton:
            inputFile = au.expandFilename(config['facilityEquipTmpl'], config)
            globalDF = pd.read_excel(inputFile, sheet_name="Global_Simulation_Parameters", header=None)
            globalDF.set_index(0, inplace=True) # spreadsheet comes in with param names in col 0.  Make these the index
            for colName, keyName in Globals.GlobalRenames.items():
                try:
                    Globals.GlobalSingleton[keyName] = globalDF.loc[colName][1]
                    Globals.GlobalSingleton[keyName] = Globals.GlobalFuncs.get(keyName, lambda x:x)(Globals.GlobalSingleton[keyName])
                except KeyError:
                    pass
            with open(configDir/userConfigFile, 'w') as ucFile:             # write userConfig.json to...
                json.dump(Globals.GlobalSingleton, ucFile, indent=4)        # ...document what's in spreadsheet
        return Globals.GlobalSingleton

# Makes a singleton pandas dataframe that contains all input values from "Facility_List" sheet of workbook
class FacilityList:
    FacilityListSingleton = None

    def getFacilityList(self, file):
        if FacilityList.FacilityListSingleton is None:
            FacilityList.FacilityListSingleton = pd.read_excel(file, sheet_name="Facility_List")
        return FacilityList.FacilityListSingleton

# Makes a singleton pandas dataframe that contains all input values from "InternalUseOnly_Validation" sheet of workbook
class InputKeys:
    InputKeysSingleton = None

    def getInputKeys(self, file):
        if InputKeys.InputKeysSingleton is None:
            InputKeys.InputKeysSingleton = pd.read_excel(file, sheet_name="InternalUseOnly_Validation")
        return InputKeys.InputKeysSingleton

# Makes a singleton dictionary for user values on each sheet of the input workbook
class UserValues:
    UserValuesSingleton = {}

    def getUserValues(self, equipmentType, file, sheet):
        if equipmentType not in UserValues.UserValuesSingleton:
            userValuesDF = pd.read_excel(file, sheet)
            userValuesDF = userValuesDF.where((pd.notnull(userValuesDF)), None)
            UserValues.UserValuesSingleton[equipmentType] = userValuesDF
        return UserValues.UserValuesSingleton[equipmentType]

# Maps equipment names to worksheet names in workbook
def mapEquipmentToSheet(equipmentType):
    equipToSheet = {"compressors": "Compressors", "dehydrators": "Dehydrators", "otherEquipment": "Other_Equipment",
                    "wellpad_wells": "Wells", "wellpad_tankflash": "Wellpad_TankFlash",
                    "wellpad_pneumatics": "Wellpad_Pneumatics",
                    "wellpad_chemicalinjectionpump": "Wellpad_ChemicalInjectionPump",
                    "composition": "Composition",
                    "facility": "Facility_List",
                    "wellpad_unloading_timing": "Wellpad_UnloadingTiming",
                    }
    return equipToSheet[equipmentType]

# Expands and cleans any nested keys
def cleanupNestedFields(intakeData):
    for singleRow in intakeData:
        expanded = {}
        deletedKeys = []
        for key in singleRow:
            if "." in key:
                splitKey = key.split(".")
                if len(splitKey) == 2:
                    mainKey, nestedKey = [splitKey[0], splitKey[1]]
                    if mainKey not in expanded:
                        expanded[mainKey] = {}
                    expanded[mainKey][nestedKey] = singleRow[key]
                    deletedKeys.append(key)
                elif len(splitKey) == 3:
                    mainKey, nestedKey1, nestedKey2 = [splitKey[0], splitKey[1], splitKey[2]]
                    if mainKey not in expanded:
                        expanded[mainKey] = {}
                    if nestedKey1 not in expanded[mainKey]:
                        expanded[mainKey][nestedKey1] = {}
                    expanded[mainKey][nestedKey1][nestedKey2] = singleRow[key]
                    deletedKeys.append(key)
        for key in deletedKeys:
            del singleRow[key]

        singleRow.update(expanded)
    return intakeData

# Matches user-facing data source names with internal shorthand names
def cleanupDataSources(intakeData):
    compressorDataSources = {"Gathering Compressor Station Study (CSU)": "GEF",
                             "Transmission and Storage Compressor Station Study (CSU)": "TS",
                             "U.S. EPA Greenhouse Gas Reporting Program (GHGRP)": "GHGRP",
                             "User specified": "User"
                            }

    leakDataSources = {"Gathering Compressor Station Study (CSU)": "GEF",
                       "Transmission and Storage Compressor Station Study (CSU)": "TS",
                       "U.S. EPA Greenhouse Gas Reporting Program (GHGRP)": "GHGRP",
                       "User specified": "User"
                      }

    leakSurveyMethods = {"OGI, Method 21, Infrared, or Acoustic": "Other Methods",
                         "Method 21 as specified in §60.5397a": "Method (a)(7)"}

    # TODO, i may have renamed somewhere...
    pneumaticDataSources = {"U.S. EPA Greenhouse Gas Reporting Program (GHGRP)": "EPA",
                            "Allen 15": "Allen15",
                            "Allen 15 ykcut": "Allen15_ykcut"
                           }

    for singleRow in intakeData:
        if "compressorDataSource" in singleRow:
            singleRow["compressorDataSource"] = compressorDataSources[singleRow["compressorDataSource"]]
        if "surveyMethod" in singleRow:
            singleRow['surveyMethod'] = leakSurveyMethods[singleRow["surveyMethod"]]
        if "reportingYear" in singleRow:
            singleRow['reportingYear'] = int(singleRow['reportingYear'])
        try:
            singleRow["leakDataSource"] = leakDataSources[singleRow["leakDataSource"]]
            singleRow["pneumaticDataSource"] = pneumaticDataSources[singleRow["pneumaticDataSource"]]
        except KeyError:
            pass

    return intakeData

# Translates long form field names to shorthand used by picker and scales data values as appropriate
def cleanupCompressorData(intakeData):

    driverTypes = {"Two stroke lean burn": "2SLB", "Four stroke lean burn": "4SLB", "Four stroke rich burn": "4SRB",
                   "Turbine": "Turbine", "Electric motor": "Electric motor"
                   }
    combustionEmissionRef = {"EPA AP-42": "AP-42", "40 CFR Subpart C": "Subpart_C"}
    for singleRow in intakeData:
        # Fix up seal type to match expected internal values
        singleRow["driverData"]["sealType"] = fixSealType(singleRow["driverData"]["sealType"])
        # Convert driver type, if given
        if singleRow["driverData"]["driverType"]:
            singleRow["driverData"]["driverType"] = driverTypes[singleRow["driverData"]["driverType"]]
        # Convert fuel consumption to efficiency, if given
        if singleRow["driverData"]["efficiency"]:
            singleRow["driverData"]["efficiency"] = u.btuPerHpHrToEfficiency(singleRow["driverData"]["efficiency"])
        singleRow["combustionEmissionReference"] = combustionEmissionRef[singleRow["combustionEmissionReference"]]
        singleRow["driverData"]["loading"] = makeLoadingDistribution(singleRow["driverData"]["loading"]["mu"],
                                                                     singleRow["driverData"]["loading"]["sigma"])
        singleRow = setCompressorTiming(singleRow)
        singleRow["operationData"]["transitionProbabilities"] = DEFAULT_XITION_PROB_MATRIX # Hard-coding this to default for now because picker will automatically adjust as needed
    return intakeData

def fixSealType(sealType):
    if sealType == "Dry":
        sealType = "drySeal"
    elif sealType == "Wet":
        sealType = "wetSeal"
    return sealType

def makeLoadingDistribution(mu, sigma):
    # If average loading was not specified, write out 'None'. The default loading will be applied in the picker.
    if not mu:
        loading = None
    # If standard deviation not given, return constant with percentage converted to per unit
    elif mu and not sigma:
        loading = mu/100
    # Otherwise make normal distribution with given mean and standard deviation
    else:
        loading = {"distribution": "normal", "mu": mu/100, "sigma": sigma/100}
    return loading

def stringToList(string):
    li = list(string.split(","))
    li = [int(x) for x in li]
    return li

def setCompressorTiming(row):
    operationData = row["operationData"]
    # If users provided times in hours, convert to percentages
    if operationData['uptime'] is not None:
        stateTimes = {'uptime': operationData['uptime'], 'percentNOP': operationData['percentNOP'], 'percentNOD': operationData['percentNOD']}
        totalTime = sum(stateTimes.values())
        if totalTime == 0:
            facID = row['facilityID']
            compID = row['compressorID']
            raise SystemExit(f"Incorrect run mode hours for facility {facID} compressor {compID}!\n"
                             f"OP, NOP, and NOD Mode Hours per Year cannot all be zero.")
        # Convert all times from hours per year to percentage
        for key, value in stateTimes.items():
            operationData[key] = u.safeDivide(value, totalTime, 0)

    # If user has entered timing data, check that it is valid and format it. Otherwise use default
    # For a constant (integer or float), pass value
    if isinstance(operationData["opTime"], int) or isinstance(operationData["opTime"], float):
        pass
    # If a list is provided, convert to distribution with min/mean/max
    elif isinstance(operationData["opTime"], str):
        operationData["opTime"] = stringToList(operationData["opTime"])
        operationData["opTime"] = checkTiming(row=row, timeList=operationData["opTime"], mode="OP")
        opTime_min  = operationData["opTime"][0]
        opTime_mean = operationData["opTime"][1]
        opTime_max  = operationData["opTime"][2]
        operationData["opTime"] = {"distribution": "triangular", "min": opTime_min, "mean": opTime_mean, "max": opTime_max}
    # Otherwise use defaults
    else:
        operationData["opTime"] = DEFAULT_TIMING["opTime"]

    if isinstance(operationData["nopTime"], int) or isinstance(operationData["nopTime"], float):
        pass
    elif isinstance(operationData["nopTime"], str):
        operationData["nopTime"] = stringToList(operationData["nopTime"])
        operationData["nopTime"] = checkTiming(row=row, timeList=operationData["nopTime"], mode="NOP")
        nopTime_min  = operationData["nopTime"][0]
        nopTime_mean = operationData["nopTime"][1]
        nopTime_max  = operationData["nopTime"][2]
        operationData["nopTime"] = {"distribution": "triangular", "min": nopTime_min, "mean": nopTime_mean, "max": nopTime_max}
    else:
        operationData["nopTime"] = DEFAULT_TIMING["nopTime"]

    if isinstance(operationData["nodTime"], int) or isinstance(operationData["nodTime"], float):
        pass
    elif isinstance(operationData["nodTime"], str):
        operationData["nodTime"] = stringToList(operationData["nodTime"])
        operationData["nodTime"] = checkTiming(row=row, timeList=operationData["nodTime"], mode="NOD")
        nodTime_min  = operationData["nodTime"][0]
        nodTime_max  = operationData["nodTime"][1]
        operationData["nodTime"] = {"distribution": "triangular", "min": nodTime_min, "max": nodTime_max}
    else:
        operationData["nodTime"] = DEFAULT_TIMING["nodTime"]
    row['operationData'].update(operationData)
    return row

# Converts timing values to list and checks that values have been entered correctly
def checkTiming(row, timeList, mode):
    compID = row['compressorID']
    facID = row['facilityID']
    for value in range(len(timeList)):
        timeList[value] = float(timeList[value])
    # Check that there are the correct number of values for each mode and that they are correctly ordered
    if mode == 'OP' or mode == 'NOP':
        if len(timeList) != 3:
            raise SystemExit(f"Incorrect parameters for {mode} mode for facility {facID} compressor {compID}!\n"
                             f"Please enter as [min, mean, max] or leave blank to use default values.")
        if timeList[1] < timeList[0]:
            raise SystemExit(f"Error! Mean is less than min for {mode} mode for facility {facID} compressor {compID}")
        if timeList[1] > timeList[2]:
            raise SystemExit(f"Error! Mean is greater than max for {mode} mode for facility {facID} compressor {compID}")
    if mode == 'NOD':
        if len(timeList) != 2:
            raise SystemExit(f"Incorrect parameters for {mode} mode for facility {facID} compressor {compID}!\n"
                             f"Please enter as [min, max] or leave blank to use default values.")
    if timeList[0] > timeList[1]:
        raise SystemExit(f"Error! Min is greater than max for {mode} mode for facility {facID} compressor {compID}")
    return timeList

# Fixes user entry which capitalizes type to internal variable which is lower case
def cleanupDehydratorData(intakeData):
    mapType = {"≥ 0.4 MMSCF/day": "large", "< 0.4 MMSCF/day": "small"}
    for row in intakeData:
        row['dehydratorType'] = mapType[row['dehydratorType']]
    return intakeData

class InputMain:
    FacilityList = FacilityList()
    InputKeys = InputKeys()
    UserValues = UserValues()
    WellList = {}

    def __init__(self, siteFilename, facilityInfo, simulationStartDate):
        self.file = siteFilename
        self.facilityInfo = facilityInfo
        self.facilityInfo = self.FacilityList.getFacilityList(self.file)
        self.simulationStartDate = datetime.datetime.strptime(simulationStartDate, '%Y-%m-%d')
        pass

    def loadIntakeData(self, equipmentType, assignFaclityInfo = True):
        self.equipmentType = equipmentType
        self.equipmentSheet = mapEquipmentToSheet(self.equipmentType)
        self.inputKeys = self.InputKeys.getInputKeys(self.file)
        self.userValues = self.UserValues.getUserValues(self.equipmentType, self.file, self.equipmentSheet)
        self.columnToKeyMap = self.mapColumnsToKeys(self.equipmentSheet)
        intakeData = self.populateIntakeData(self.equipmentType, assignFaclityInfo)
        intakeData = self.cleanupIntakeData(intakeData, self.equipmentType)
        return intakeData

    # Returns a dictionary that maps excel column headers to intake data keys
    def mapColumnsToKeys(self, sheet):
        excelColumns = list(self.inputKeys.loc[self.inputKeys["Sheet"] == sheet]["Excel Column"])
        intakeDataKeys = list(self.inputKeys.loc[self.inputKeys["Sheet"] == sheet]["Intake Data Key"])
        columnToKeyMap = dict(zip(excelColumns, intakeDataKeys))
        return columnToKeyMap

    # Populates the intake data dictionary from a specific worksheet and file
    def populateIntakeData(self, equipmentType, attachFacilityInfo):
        intakeList = []
        intakeDict = {}
        facilityList = self.assignFacilityParams()
        for column, values in self.userValues.iteritems():
            # sanitize column name:
            # multiple consecutive white space goes to single white space (" ")
            column_clean = re.sub('[ \t\n]+', ' ', column.strip())
            if column in self.columnToKeyMap or column_clean in self.columnToKeyMap:
                try:
                    intakeDataKey = self.columnToKeyMap[column]
                except KeyError:
                    intakeDataKey = self.columnToKeyMap[column_clean]
                intakeDict[intakeDataKey] = list(values)
        for n in range(len(self.userValues)):
            singleEquipment = {}
            for key in intakeDict.keys():
                singleEquipment[key] = intakeDict[key][n]
            intakeList.append(singleEquipment)
        if not attachFacilityInfo:
            return intakeList
        for singleFacility in facilityList:
            facilityID = singleFacility['facilityID']
            for singleEquipment in intakeList:
                # Saves wells off as a separate dictionary
                if equipmentType == 'wellpad_wells':
                    thisWell = copy.deepcopy(singleEquipment)
                    InputMain.WellList[singleEquipment['wellID']] = thisWell
                if singleEquipment['facilityID'] == facilityID:
                    singleEquipment.update(singleFacility)
                # If equipment is associated with a specific well use the gas composition profile from the well
                wellID = singleEquipment.get('wellID', None)
                if wellID:
                    wellID = singleEquipment['wellID']
                    well = InputMain.WellList[wellID]
                    singleEquipment['profile_id'] = well['profile_id']
        return intakeList
    # Cleans up, scales, and pre-processes intake data for use by pickers
    def cleanupIntakeData(self, intakeData, equipmentType):
        intakeData = cleanupNestedFields(intakeData)
        intakeData = cleanupDataSources(intakeData)
        if equipmentType == "compressors":
            intakeData = cleanupCompressorData(intakeData)
        elif equipmentType == "dehydrators":
            intakeData = cleanupDehydratorData(intakeData)
        elif equipmentType == 'wellpad_unloading_timing':
            intakeData = self.cleanupUnloadingTiming(intakeData)
        elif equipmentType == 'wellpad_wells':
            intakeData = self.cleanupWellData(intakeData)

        return intakeData

    # Gets the facility parameters as defined by the facility list
    def assignFacilityParams(self):
        facilityList = []
        facilityKeys = self.mapColumnsToKeys("Facility_List")
        facilitySet = set(self.facilityInfo['Facility_ID'])
        facilityDict = {}
        for column, values in self.facilityInfo.iteritems():
            facilityKey = facilityKeys[column]
            facilityDict[facilityKey] = list(values)
        for facility in range(len(facilitySet)):
            singleFacility = {}
            for key in facilityDict.keys():
                singleFacility[key] = facilityDict[key][facility]
            facilityList.append(singleFacility)

        return facilityList

    def datetimeToDEStime(self, dt):
        """take Series of time stamp, convert it to DES timestamp"""
        ts = (dt - self.simulationStartDate).total_seconds()
        return ts

    def cleanupUnloadingTiming(self, intakeData):

        for row in intakeData:
            row['venting_start_timestamp'] = self.datetimeToDEStime(row['venting_start_datetime'])
        return intakeData

    def cleanupWellData(self, intakeData):
        for row in intakeData:
            if isinstance(row['basin_id_aapg'], numbers.Number):
                row['basin_id_aapg'] = '%d' % int(row['basin_id_aapg'])
        return intakeData

